﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Core;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.Variants;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {


  public class ClusterAlgorithm<C,CO> where C : Cluster<CO> where CO : ClusterElement {

    //public static List<C> KMeans(IVRPProblemInstance instance, IRandom random, List<CO> clusterObjects, int k, double changeThreshold) {
    //  HashSet<int> initMeans = new HashSet<int>();
    //  int nextMean = -1;

    //  for (int i = 0; i < k && i < clusterObjects.Count; i++) {
    //    //C cluster = new C()
    //  }




    //}

    //private static int KMeansRun(List<C> clusters, List<CO> clusterObjects) {
      
    //}

  }

  #region Cluster
  public interface ICluster<T> where T : IClusterElement {
    double CalculateImpact(T o);
    void CalculateMean();
    void SetMean(T o);
    bool AddObject(T o);
  }
  public abstract class Cluster<T> : ICluster<T> where T : ClusterElement {
    public int Id;
    protected readonly IVRPProblemInstance instance;
    public List<T> Elements { get; private set; }

    public T Mean { get; set; }

    public double Variance { get; set; }

    public Cluster(IVRPProblemInstance instance, int id) {
      this.instance = instance;
      Id = id;
      Elements = new List<T>();

    }
    public bool AddObject(T o) {
      Elements.Add(o);

      bool clusterChanged = o.ClusterId != Id;
      o.ClusterId = Id;
      return clusterChanged;
    }

    public void SetMean(T o) {
      Mean = o;
      Mean.ClusterId = Id;
    }

    public abstract void CalculateMean();

    public abstract void CalculateVariance();

    public abstract double CalculateImpact(T o);

  }
  public class SpatialDistanceCluster : Cluster<SpatialDistanceClusterElement> {
    public SpatialDistanceCluster(IVRPProblemInstance instance, int id) : base(instance, id) {
    }

    public override void CalculateMean() {
      int dimensions = instance.Coordinates.Columns;
      SpatialDistanceClusterElement mean = new SpatialDistanceClusterElement(-1, dimensions, Id);
      foreach (SpatialDistanceClusterElement o in Elements) {
        double[] coordinates = instance.GetCoordinates(o.Id);
        for (int i = 0; i < coordinates.Length; i++) {
          mean.Coordinates[i] += (coordinates[i] / Elements.Count);
        }
      }
      Mean = mean;
    }

    public override void CalculateVariance() {
      if(Mean == null)
        CalculateMean();

      Variance = 0.0;
      foreach (SpatialDistanceClusterElement o in Elements) {
        Variance += Math.Pow(Mean.GetDistance(o), 2);
      }
      Variance /= Elements.Count;
    }

    public override double CalculateImpact(SpatialDistanceClusterElement o) {
      if (Mean == null)
        CalculateMean();

      double newVariance = (Variance*Elements.Count + Math.Pow(Mean.GetDistance(o), 2))/(Elements.Count + 1);
      return newVariance - Variance;
    }
  }

  public class TemporalDistanceCluster : Cluster<TemporalDistanceClusterElement> {
    public TemporalDistanceCluster(ITimeWindowedProblemInstance instance, int id) : base(instance, id) {
    }


    public override void CalculateMean() {
      TemporalDistanceClusterElement mean = new TemporalDistanceClusterElement(-1, Id);
      foreach (TemporalDistanceClusterElement e in Elements) {
        mean.ReadyTime += e.ReadyTime/Elements.Count;
        mean.DueTime += e.DueTime/Elements.Count;
      }
      Mean = mean;
    }

    public override void CalculateVariance() {
      if (Mean == null)
        CalculateMean();

      Variance = 0.0;
      foreach (TemporalDistanceClusterElement e in Elements) {
        Variance += Math.Pow(Math.Abs(Mean.ReadyTime - e.ReadyTime) + Math.Abs(Mean.DueTime - e.DueTime), 2);
      }
      Variance /= Elements.Count;
    }

    public override double CalculateImpact(TemporalDistanceClusterElement o) {
      if(Mean == null)
        CalculateMean();

      double newVariance = (Variance * Elements.Count + Math.Pow(Math.Abs(Mean.ReadyTime - o.ReadyTime) + Math.Abs(Mean.DueTime - o.DueTime), 2)) / (Elements.Count + 1);
      return newVariance - Variance;
    }
  }
  #endregion Cluster


  #region ClusterElement
  public interface IClusterElement {
    double GetDistance(IClusterElement o);
  }
  public abstract class ClusterElement : IClusterElement {
    public int Id { get; set; }
    public int ClusterId { get; set; }

    private ClusterElement() { }

    public ClusterElement(int id, int clusterId = -1) {
      Id = id;
      ClusterId = clusterId;
    }

    public abstract double GetDistance(IClusterElement o);
  }
  public class SpatialDistanceClusterElement : ClusterElement {

    public double[] Coordinates { get; set; }

    public SpatialDistanceClusterElement(int id, int clusterId = -1) : base(id, clusterId) {
    }

    public SpatialDistanceClusterElement(int id, double[] coordinates, int clusterId = -1) : base(id, clusterId) {
      Coordinates = coordinates;
    }

    public SpatialDistanceClusterElement(int id, int dimensions, int clusterId = -1) : base(id, clusterId) { 
      Coordinates = new double[dimensions];
    }

    public override double GetDistance(IClusterElement o) {
      var other = o as SpatialDistanceClusterElement;
      if (other == null) {
        throw new ArgumentException(string.Format("Distance calculation between {0} and {1} is not possible.",
          typeof(SpatialDistanceClusterElement), o.GetType()));
      }

      double distance = 0.0;
      for (int i = 0; i < other.Coordinates.Length; i++) {
        distance += Math.Pow(Coordinates[i] - other.Coordinates[i], 2);
      }

      return Math.Sqrt(distance);
    }
  }

  public class TemporalDistanceClusterElement : ClusterElement {
    public double ReadyTime { get; set; }
    public double DueTime { get; set; }

    public TemporalDistanceClusterElement(int id, int clusterId = -1) : base(id, clusterId) {
    }

    public TemporalDistanceClusterElement(int id, double readyTime, double dueTime, int clusterId = -1) : base(id, clusterId) {
      ReadyTime = readyTime;
      DueTime = dueTime;
    }

    public override double GetDistance(IClusterElement o) {
      var other = o as TemporalDistanceClusterElement;
      if (other == null) {
        throw new ArgumentException(string.Format("Distance calculation between {0} and {1} is not possible.",
          typeof(TemporalDistanceClusterElement), o.GetType()));
      }

      double distance = 0.0;
      distance += Math.Abs(ReadyTime - other.ReadyTime);
      distance += Math.Abs(DueTime - other.DueTime);

      return distance;

      //return GetDistanceEuklid(other);
    }

    public double GetDistanceEuklid(TemporalDistanceClusterElement other) {
      double distance = 0.0;
      distance += Math.Pow(ReadyTime - other.ReadyTime, 2);
      distance += Math.Pow(DueTime - other.DueTime, 2);

      return Math.Sqrt(distance);
    }

  }

  #endregion ClusterElement
}
