﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("GeographicDistanceClusterCreator", "Creates a VRP solution by clustering customers first with a KMeans-algorithm based on their geographic position and building tours afterwards alternatevly in a random or a greedy fashion.")]
  [StorableClass]
  public sealed class GeographicDistanceClusterCreator : ClusterCreator {

    [StorableConstructor]
    private GeographicDistanceClusterCreator(bool deserializing) : base(deserializing) { }

    public GeographicDistanceClusterCreator() : base() {
    }

    private GeographicDistanceClusterCreator(GeographicDistanceClusterCreator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GeographicDistanceClusterCreator(this, cloner);
    }

    public static List<SpatialDistanceClusterElement> CreateClusterObjects(IVRPProblemInstance instance) {
      IPickupAndDeliveryProblemInstance pdp = instance as IPickupAndDeliveryProblemInstance;

      // add all customers
      List<int> customers = new List<int>();
      for (int i = 1; i <= instance.Cities.Value; i++) {
        if (pdp == null || pdp.GetDemand(i) >= 0)
          customers.Add(i);
      }

      // wrap stops in SpatialDistanceClusterElement
      List<SpatialDistanceClusterElement> clusterObjects = new List<SpatialDistanceClusterElement>();
      foreach (int customer in customers) {
        clusterObjects.Add(new SpatialDistanceClusterElement(customer, instance.GetCoordinates(customer)));
      }
      return clusterObjects;
    }

    public static PotvinEncoding CreateSolution(IVRPProblemInstance instance, IRandom random, int minK, int maxK, double clusterChangeThreshold, int creationOption) {
      PotvinEncoding result = new PotvinEncoding(instance);

      List<SpatialDistanceClusterElement> clusterObjects = CreateClusterObjects(instance);

      int k = random.Next(minK, maxK);
      List<SpatialDistanceCluster> clusters = KMeans(instance, random, clusterObjects, k, clusterChangeThreshold);

      // (3) build tours
      // (a) shuffle
      // (b) greedy
      foreach (var c in clusters) {
        Tour newTour = new Tour();
        result.Tours.Add(newTour);

        if (creationOption == 0) {
          // (a) shuffle
          c.Elements.Shuffle(random);
          foreach (var o in c.Elements) {
            newTour.Stops.Add(o.Id);
          }
        } else {
          // (b) greedy
          foreach (var o in c.Elements) {
            newTour.Stops.Add(o.Id);
          }
          GreedyTourCreation(instance, result, newTour, false);
        }
      }

      return result;
    }

    private static List<SpatialDistanceCluster> KMeans(IVRPProblemInstance instance, IRandom random, List<SpatialDistanceClusterElement> objects, int k, double changeTreshold) {

      List<SpatialDistanceCluster> clusters = new List<SpatialDistanceCluster>();
      HashSet<int> initMeans = new HashSet<int>();
      int nextMean = -1;

      // (1) initialize each cluster with a random first object (i.e. mean)
      for (int i = 0; i < k && i < objects.Count; i++) {
        SpatialDistanceCluster cluster = new SpatialDistanceCluster(instance, i);

        do {
          nextMean = random.Next(0, objects.Count);
        } while (initMeans.Contains(nextMean));
        initMeans.Add(nextMean);
        cluster.SetMean(objects[nextMean]);
        clusters.Add(cluster);
      }

      // (2) repeat clustering until change rate is below threshold
      int changes = 0;
      double changeRate = 1.0;

      do {
        changes = KMeansRun(clusters, objects);
        changeRate = changes / objects.Count;

      } while (changeRate > changeTreshold);

      // remove empty clusters
      clusters.RemoveAll(c => c.Elements.Count.Equals(0));

      return clusters;
    }

    private static int KMeansRun(List<SpatialDistanceCluster> clusters, List<SpatialDistanceClusterElement> objects) {
      int changes = 0;

      foreach (var c in clusters) {
        c.Elements.Clear();
      }

      foreach (var o in objects) {
        int bestClusterIdx = 0;
        double minImpact = clusters[0].CalculateImpact(o);
        for (int i = 1; i < clusters.Count; i++) {
          double impact = clusters[i].CalculateImpact(o);
          if (impact < minImpact) {
            minImpact = impact;
            bestClusterIdx = i;
          }
        }
        if (clusters[bestClusterIdx].AddObject(o))
          changes++;
      }

      foreach (var c in clusters) {
        c.CalculateMean();
        c.CalculateVariance();
      }

      return changes;
    }

    public override IOperation InstrumentedApply() {
      IRandom random = RandomParameter.ActualValue;

      int minK = (MinK.Value.Value > 0) ? MinK.Value.Value : 1;
      int maxK = (MaxK.Value != null) ? MaxK.Value.Value : ProblemInstance.Vehicles.Value;
      double clusterChangeThreshold = (ClusterChangeThreshold.Value.Value >= 0.0 &&
                                       ClusterChangeThreshold.Value.Value <= 1.0)
        ? ClusterChangeThreshold.Value.Value
        : 0.0;

      // normalize probabilities
      double max = TourCreationProbabilities.Value.Max();
      double[] probabilites = new double[2];
      for (int i = 0; i < TourCreationProbabilities.Value.Length; i++) {
        probabilites[i] = TourCreationProbabilities.Value[i] / max;
      }

      List<int> creationOptions = new List<int>() { 0, 1 };
      int creationOption = creationOptions.SampleProportional(random, 1, probabilites, false, false).First();

      VRPToursParameter.ActualValue = CreateSolution(ProblemInstance, random, minK, maxK, clusterChangeThreshold, creationOption);
      return base.InstrumentedApply();
    }
  }
}
